# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .workspace_display_config import WorkspaceDisplayConfig


class WorkspaceRead(UniversalBaseModel):
    id: str
    name: str = pydantic.Field()
    """
    The name of the Workspace.
    """

    label: str
    created: dt.datetime
    display_config: typing.Optional[WorkspaceDisplayConfig] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
